Type TBall
	' stores all balls
	Global balls:TBall[]
	' graphics
	Field imgLight:TImage
	Field imgBall:TImage
	Field imgRefl:TImage
	' size
	Field rad:Double
	' position
	Field angle:Double, dist0:Double, dist1:Double
	Field distf:Double, distg:Double	' lift and hard lift value respectively
	Field lifting:Int, liftAmount:Int	' lift data
	Field liftSTime:Int, liftETime:Int
	' rotation
	Field dir:Double
	Field rot:Double
	Field rotVel:Double, addRotVel:Double	' peripheral speed in px/s
	Field slipping:Int, slipAmount:Int		' slip data
	Field slipSTime:Int
	' color
	Field red:Int,grn:Int,blu:Int
	
	' create ball
	Function Create:TBall(pAngle:Double, pDist:Double, pRad:Double, pDir:Double)
		Local ball:TBall = New TBall
		ball.rad		= pRad
		ball.angle	= pAngle
		ball.dist0	= pDist
		ball.dist1	= pDist+1+pRad*2
		ball.dir		= pDir
		ball.red		= 128
		ball.grn		= 128
		ball.blu		= 128
		ball.GenerateImages()
		balls :+ [ball]
		Return ball
	End Function
	
	' lift ball slow
	Method Lift(pTime:Int, pAmount:Int)
		lifting		= True
		liftAmount	= pAmount
		liftSTime		= TDemo.playTime
		liftETime		= pTime
	End Method
	' hit ball immediately
	Method Hit()
		lifting = False
	End Method
	' start to slip
	Method Slip(pAmount:Int)
		slipping		= True
		slipAmount	= pAmount
		slipSTime		= TDemo.playTime
	End Method
	
	Method Color(pRed:Int, pGrn:Int, pBlu:Int)
		red = pRed
		grn = pGrn
		blu = pBlu
	End Method
	
	Function DrawBalls()
		For Local ball:TBall = EachIn balls
			ball.Draw()
		Next
	End Function
	
	Method Draw()
		' calculate position
		If lifting
			Local dt:Double	= TDemo.playTime -liftSTime
			Local ttot:Double	= liftETime -liftSTime
			distf = (dt/ttot)^2 *liftAmount/127.0
		Else
			distf = 0
		EndIf
		Local d:Double = dist0 +(dist1 -dist0)*Max(distf, distg)
		Local x:Double = 400 +Cos(angle)*d
		Local y:Double = 300 +Sin(angle)*d
		' calculate rotation
		If slipping
			Local dt:Double	= TDemo.playTime -slipSTime
			Local ttot:Double	= 1000
			Local arg:Double	= (dt/ttot)^2
			rotVel = (0.5+0.5*Exp(-arg))*(1-Exp(-50*arg)) *2 *slipAmount
		Else
			rotVel = 0
		EndIf
		rot :+ ((TDemo.rotVel +rotVel +addRotVel)/rad)/Pi*0.18*dir *TDemo.dTime

		SetScale 1,1
		SetAlpha 1
		SetColor red*TDemo.lightRed, grn*TDemo.lightGrn, blu*TDemo.lightBlu
		SetBlend ALPHABLEND
		SetRotation -90
		DrawImage imgLight, x,y	
	
		SetColor 255,255,255
		SetBlend SHADEBLEND
		SetRotation(rot)
		DrawImage imgBall, x,y
	
		SetColor 64*TDemo.lightRed,64*TDemo.lightGrn,64*TDemo.lightBlu
		SetBlend LIGHTBLEND
		SetRotation -90
		DrawImage imgRefl, x,y
	End Method
	
	Method GenerateImages()
		Local size:Int = rad*2+3
		Local half:Int = rad+1
		imgBall = CreateImage(size,size,1,FILTEREDIMAGE)
		MidHandleImage(imgBall)
		Local pm:TPixmap = LockImage(imgBall)
		For Local y:Int = 0 Until size
			For Local x:Int = 0 Until size
				Local dx:Double = half-x
				Local dy:Double = half-y
				Local phi:Double = ATan2(dy,dx)
				Local dist:Double = Sqr(dx*dx+dy*dy)
				Local f:Double
				Local gr:Int = 255
				Local n:Int = 10+rad/16
				Local g:Double = 0.7-rad/240
				f=0.8
				For Local k:Int = 1 To 10
					f :+ (1-Cos(dist/rad*90)*0.8)*0.2*Sin((phi+dist*g)*(2*k-1)*n)/Double(2*k-1)
				Next
				f = 1 -((dist<rad)+(dist<=rad+1 And dist>=rad)*(rad+1-dist))*(1-f)
				gr :* f
				gr = Max(Min(255,gr),0)
				WritePixel(pm, x,y, (255 Shl 24)|(gr Shl 16)|(gr Shl 8)|gr)
			Next
		Next
		UnlockImage(imgBall)
		
		imgRefl = CreateImage(size,size)
		MidHandleImage(imgRefl)
		pm = LockImage(imgRefl)
		For Local y:Int = 0 Until size
			For Local x:Int = 0 Until size
				Local dx:Double = half-x
				Local dy:Double = half-y
				Local phi:Double = ATan2(dy,dx)
				Local dist:Double = Sqr(dx*dx+dy*dy)
				Local dmx:Double = half*1.0625+Sqr((half*0.8125)^2-(dy^2))
				Local dmn:Double = half*2.5-Sqr((half*1.5625)^2-(dy^2))
				Local f:Double
				Local gr:Int = 255
				If dmx=0/0.0 Then dmx=0
				' light spot
				f = (x<=dmx)+(x<=dmx+1 And x>dmx)*(dmx+1-x)
				f :* (x>=dmn)+(x>=dmn-1 And x<dmn)*(x-dmn+1)
				f :* 0.9
				' gradient
				f :+ 0.1*(x/Double(size)*0.2+Cos(dist/Double(half)*90.0)^0.8*0.8)
				' mask
				f :* (dist<rad)+(dist<=rad+1 And dist>=rad)*(rad+1-dist)
				gr :* f
				gr=Max(Min(255,gr),0)
				WritePixel(pm, x,y, (255 Shl 24)|(gr Shl 16)|(gr Shl 8)|gr)
			Next
		Next
		UnlockImage(imgRefl)
		
		imgLight = CreateImage(size,size)
		MidHandleImage(imgLight)
		pm = LockImage(imgLight)
		For Local y:Int = 0 Until size
			For Local x:Int = 0 Until size
				Local dx:Double = half-x
				Local dy:Double = half-y
				Local phi:Double = ATan2(dy,dx)
				Local dist:Double = Sqr(dx*dx+dy*dy)
				Local dmx:Double = half*1.0625+Sqr((half*0.8125)^2-(dy^2))
				Local dmn:Double = half*2.5-Sqr((half*1.5625)^2-(dy^2))
				Local f:Double
				Local gr:Int = 255
				Local a:Int = 255
				If dmx=0/0.0 Then dmx=0
				' light spot
				f = (x<=dmx)+(x<=dmx+1 And x>dmx)*(dmx+1-x)
				f :* (x>=dmn)+(x>=dmn-1 And x<dmn)*(x-dmn+1)
				f :* 0.1
				' gradient
				f :+ 0.9*(x/Double(size)*0.6+Cos(dist/Double(half)*90.0)^0.8*0.9)
				' ambient light
				f = 0.1 +0.9*f
				gr :* f
				' mask
				f = (dist<rad)+(dist<=rad+1 And dist>=rad)*(rad+1-dist)
				a :* f
				gr=Max(Min(255,gr),0)
				a=Max(Min(255,a),0)
				WritePixel(pm, x,y, (a Shl 24)|(gr Shl 16)|(gr Shl 8)|gr)
			Next
		Next
		UnlockImage(imgLight)
	End Method
End Type

