' controller (one controller per used channel)
Type TController Abstract
	' midi CC types
	Const CC_COLORRED:Int	= 110
	Const CC_COLORGRN:Int	= 111
	Const CC_COLORBLU:Int	= 112
	Const CC_MODE:Int		= 120
	Const CC_ROTVEL:Int		= 121
	Const CC_LIFT:Int		= 122
	Const CC_END:Int		= 127

	' channel assigned to this controller
	Field channel:Int
	' event stack (first in first out)
	Field events:TList = New TList
	
	' add event to event stack
	Method PushEvent(pEvent:TMidiEvent)
		' create controller event
		Local e:TControllerEvent = New TControllerEvent
		e.time	= TDemo.playTime +TDemo.LEAD_TIME
		e.event	= pEvent
		' push
		events.AddLast(e)
	End Method
	
	' run controller
	Method Update() Abstract
End Type

Type TControllerBall Extends TController
	' ball assigned to this controller
	Field ball:Int
	' temp values
	Field _mode:Int
	
	Function Create:TControllerBall(pChannel:Int, pBall:Int)
		Local c:TControllerBall = New TControllerBall
		c.channel	= pChannel
		c.ball	= pBall
		Return c
	End Function
	
	Method Update()
		Local first:Byte = True
		' search active event
		For Local e:TControllerEvent = EachIn events
			' midi event
			Local me:TMidiEvent = e.event
			Select me.eventType
			' note on
			Case TMidiEvent.NOTE_ON
				' lift & hit
				If _mode=0 And first
					first=False
					' auto lift
					If TDemo.playTime<e.time And e.extra=0
						TDemo.balls[ball].Lift(e.time, me.parameter2)
						e.extra = 1
					' hit
					ElseIf TDemo.playTime>=e.time
						TDemo.balls[ball].Hit()
					EndIf
				' slip
				ElseIf _mode=1 And TDemo.playTime>=e.time
					TDemo.balls[ball].Slip(me.parameter2)	
				EndIf
			' note off
			Case TMidiEvent.NOTE_OFF
				If _mode=1 And TDemo.playTime>=e.time
					TDemo.balls[ball].slipping=False
				EndIf
			' control
			Case TMidiEvent.CONTROLLER
				If TDemo.playTime>=e.time
					Select me.parameter1
					' set color
					Case CC_COLORRED
						TDemo.balls[ball].red = me.parameter2/127:Double*255
					Case CC_COLORGRN
						TDemo.balls[ball].grn = me.parameter2/127:Double*255
					Case CC_COLORBLU
						TDemo.balls[ball].blu = me.parameter2/127:Double*255
					' set mode
					Case CC_MODE
						_mode = me.parameter2
					' set additional rot velocity
					Case CC_ROTVEL
						TDemo.balls[ball].addRotVel = me.parameter2
					' set additional lift
					Case CC_LIFT
						TDemo.balls[ball].distg = me.parameter2/127:Double
					End Select
				EndIf
			End Select
			' remove triggered events
			If TDemo.playTime>=e.time
				events.Remove(e)
			EndIf
		Next
	End Method
End Type

Type TControllerBallArray Extends TController
	' balls assigned to this controller
	Field balls:Int[]
	' notes for balls
	Field notes:Int[]
	' temp values
	Field _mode:Int

	Function Create:TControllerBallArray(pChannel:Int, pBalls:Int[], pNotes:Int[])
		Local c:TControllerBallArray = New TControllerBallArray
		c.channel	= pChannel
		c.balls	= pBalls
		c.notes	= pNotes
		Return c
	End Function

	Method Update()
		Local first:Byte[] = New Byte[balls.length]
		For Local i:Int = 0 Until first.length
			first[i]=True
		Next
		' search active event
		For Local e:TControllerEvent = EachIn events
			' midi event
			Local me:TMidiEvent = e.event
			Select me.eventType
			' note on
			Case TMidiEvent.NOTE_ON
				' find ball for note
				Local i:Int = _BallForNote(me.parameter1)
				' lift & hit
				If _mode=0 And first[i]
					first[i]=False
					' auto lift
					If TDemo.playTime<e.time And e.extra=0
						TDemo.balls[balls[i]].Lift(e.time, me.parameter2)
						e.extra = 1
					' hit
					ElseIf TDemo.playTime>=e.time
						TDemo.balls[balls[i]].Hit()
					EndIf
				' slip
				ElseIf _mode=1 And TDemo.playTime>=e.time
					TDemo.balls[balls[i]].Slip(me.parameter2)	
				EndIf
			' note off
			Case TMidiEvent.NOTE_OFF
				' find ball for note
				Local i:Int = _BallForNote(me.parameter1)
				If _mode=1 And TDemo.playTime>=e.time
					TDemo.balls[balls[i]].slipping=False
				EndIf
			' control
			Case TMidiEvent.CONTROLLER
				If TDemo.playTime>=e.time
					Select me.parameter1
					' set color
					Case CC_COLORRED
						For Local i:Int = EachIn balls
							TDemo.balls[i].red = me.parameter2/127:Double*255
						Next
					Case CC_COLORGRN
						For Local i:Int = EachIn balls
							TDemo.balls[i].grn = me.parameter2/127:Double*255
						Next
					Case CC_COLORBLU
						For Local i:Int = EachIn balls
							TDemo.balls[i].blu = me.parameter2/127:Double*255
						Next
					' set mode
					Case CC_MODE
						_mode = me.parameter2
					' set additional rot velocity
					Case CC_ROTVEL
						For Local i:Int = EachIn balls
							TDemo.balls[i].addRotVel = me.parameter2
						Next
					' set additional lift
					Case CC_LIFT
						For Local i:Int = EachIn balls
							TDemo.balls[i].distg = me.parameter2/127:Double
						Next
					End Select
				EndIf
			End Select
			' remove triggered events
			If TDemo.playTime>=e.time
				events.Remove(e)
			EndIf
		Next
	End Method
	
	' find ball index for given note
	Method _BallForNote:Int(pNote:Int)
		For Local n:Int = 0 Until notes.length
			If notes[n]=pNote
				Return n
			EndIf
		Next
	End Method
End Type

Type TControllerPad Extends TController
	Function Create:TControllerPad(pChannel:Int)
		Local c:TControllerPad = New TControllerPad
		c.channel	= pChannel
		Return c
	End Function
	
	Method Update()
		' search active event
		For Local e:TControllerEvent = EachIn events
			' play when ready (no lead time nescessary)
			If TDemo.playTime>=e.time
				' midi event
				Local me:TMidiEvent = e.event
				Select me.eventType
				' play note
				Case TMidiEvent.NOTE_ON
					TDemo.background.Play(me.parameter1, me.parameter2)
				' stop note
				Case TMidiEvent.NOTE_OFF
					TDemo.background.Stop()
				' control
				Case TMidiEvent.CONTROLLER
					Select me.parameter1
					' set color
					Case CC_COLORRED
						TDemo.background.red = 255*(me.parameter2/127:Double)
					Case CC_COLORGRN
						TDemo.background.grn = 255*(me.parameter2/127:Double)
					Case CC_COLORBLU
						TDemo.background.blu = 255*(me.parameter2/127:Double)
					End Select
				End Select
				' remove triggered event
				events.Remove(e)
			EndIf
		Next
	End Method
End Type

Type TControllerMain Extends TController
	Function Create:TControllerMain(pChannel:Int)
		Local c:TControllerMain = New TControllerMain
		c.channel	= pChannel
		Return c
	End Function
	
	Method Update()
		' search active event
		For Local e:TControllerEvent = EachIn events
			' play when ready (no lead time nescessary)
			If TDemo.playTime>=e.time
				' midi event
				Local me:TMidiEvent = e.event
				Select me.eventType
				' control
				Case TMidiEvent.CONTROLLER
					Select me.parameter1
					' set color
					Case CC_COLORRED
						TDemo.lightRed = me.parameter2/127:Double
					Case CC_COLORGRN
						TDemo.lightGrn = me.parameter2/127:Double
					Case CC_COLORBLU
						TDemo.lightBlu = me.parameter2/127:Double
					' set global rotational velocity
					Case CC_ROTVEL
						TDemo.rotVel = me.parameter2
					' song end
					Case CC_END
						If me.parameter2=1
							TDemo.ShutDown()
						EndIf
					End Select
				End Select
				' remove triggered event
				events.Remove(e)
			EndIf
		Next
	End Method
End Type

' controller events
Type TControllerEvent
	Field time:Int
	Field event:TMidiEvent
	Field extra:Int
End Type

